#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country and subdivision
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

__all__ = ("DateLike", "HolidayBase", "HolidaySum")

import warnings
from datetime import date, datetime, timedelta
from typing import Any, Dict, Iterable, List, Mapping, Optional, Set, Tuple
from typing import Union, cast

from dateutil.parser import parse

from holidays.constants import SAT, SUN

DateLike = Union[date, datetime, str, float, int]


class HolidayBase(Dict[date, str]):
    """
    A dict-like object containing the holidays for a specific country (and
    province or state if so initiated); inherits the dict class (so behaves
    similarly to a dict). Dates without a key in the Holiday object are not
    holidays.

    The key of the object is the date of the holiday and the value is the name
    of the holiday itself. When passing the date as a key, the date can be
    expressed as one of the following formats:

    * datetime.datetime type;
    * datetime.date types;
    * a float representing a Unix timestamp;
    * or a string of any format (recognized by datetime.parse).

    The key is always returned as a `datetime.date` object.

    To maximize speed, the list of holidays is built as needed on the fly, one
    calendar year at a time. When you instantiate the object, it is empty, but
    the moment a key is accessed it will build that entire year's list of
    holidays. To prepopulate holidays, instantiate the class with the years
    argument:

    us_holidays = holidays.US(years=2020)

    It is generally instantiated using the :func:`country_holidays` function.

    The key of the :class:`dict`-like :class:`HolidayBase` object is the
    `date` of the holiday, and the value is the name of the holiday itself.
    Dates where a key is not present are not public holidays (or, if
    **observed** is False, days when a public holiday is observed).

    When passing the `date` as a key, the `date` can be expressed in one of the
    following types:

    * :class:`datetime.date`,
    * :class:`datetime.datetime`,
    * a :class:`str` of any format recognized by :func:`dateutil.parser.parse`,
    * or a :class:`float` or :class:`int` representing a POSIX timestamp.

    The key is always returned as a :class:`datetime.date` object.

    To maximize speed, the list of public holidays is built on the fly as
    needed, one calendar year at a time. When the object is instantiated
    without a **years** parameter, it is empty, but, unless **expand** is set
    to False, as soon as a key is accessed the class will calculate that entire
    year's list of holidays and set the keys with them.

    If you need to list the holidays as opposed to querying individual dates,
    instantiate the class with the **years** parameter.

    Example usage:

    >>> from holidays import country_holidays
    >>> us_holidays = country_holidays('US')
    # For a specific subdivisions (e.g. state or province):
    >>> calif_holidays = country_holidays('US', subdiv='CA')

    The below will cause 2015 holidays to be calculated on the fly:

    >>> from datetime import date
    >>> assert date(2015, 1, 1) in us_holidays

    This will be faster because 2015 holidays are already calculated:

    >>> assert date(2015, 1, 2) not in us_holidays

    The :class:`HolidayBase` class also recognizes strings of many formats
    and numbers representing a POSIX timestamp:

    >>> assert '2014-01-01' in us_holidays
    >>> assert '1/1/2014' in us_holidays
    >>> assert 1388597445 in us_holidays

    Show the holiday's name:

    >>> us_holidays.get('2014-01-01')
    "New Year's Day"

    Check a range:

    >>> us_holidays['2014-01-01': '2014-01-03']
    [datetime.date(2014, 1, 1)]

    List all 2020 holidays:

    >>> us_holidays = country_holidays('US', years=2020)
    >>> for day in us_holidays.items():
    ...     print(day)
    (datetime.date(2020, 1, 1), "New Year's Day")
    (datetime.date(2020, 1, 20), 'Martin Luther King Jr. Day')
    (datetime.date(2020, 2, 17), "Washington's Birthday")
    (datetime.date(2020, 5, 25), 'Memorial Day')
    (datetime.date(2020, 7, 4), 'Independence Day')
    (datetime.date(2020, 7, 3), 'Independence Day (Observed)')
    (datetime.date(2020, 9, 7), 'Labor Day')
    (datetime.date(2020, 10, 12), 'Columbus Day')
    (datetime.date(2020, 11, 11), 'Veterans Day')
    (datetime.date(2020, 11, 26), 'Thanksgiving')
    (datetime.date(2020, 12, 25), 'Christmas Day')

    Some holidays are only present in parts of a country:

    >>> us_pr_holidays = country_holidays('US', subdiv='PR')
    >>> assert '2018-01-06' not in us_holidays
    >>> assert '2018-01-06' in us_pr_holidays

    Append custom holiday dates by passing one of:

    * a :class:`dict` with date/name key/value pairs (e.g.
      ``{'2010-07-10': 'My birthday!'}``),
    * a list of dates (as a :class:`datetime.date`, :class:`datetime.datetime`,
      :class:`str`, :class:`int`, or :class:`float`); ``'Holiday'`` will be
      used as a description,
    * or a single date item (of one of the types above); ``'Holiday'`` will be
      used as a description:

    >>> custom_holidays = country_holidays('US', years=2015)
    >>> custom_holidays.update({'2015-01-01': "New Year's Day"})
    >>> custom_holidays.update(['2015-07-01', '07/04/2015'])
    >>> custom_holidays.update(date(2015, 12, 25))
    >>> assert date(2015, 1, 1) in custom_holidays
    >>> assert date(2015, 1, 2) not in custom_holidays
    >>> assert '12/25/2015' in custom_holidays

    For special (one-off) country-wide holidays handling use
    :attr:`special_holidays`:

    .. code-block:: python

        special_holidays = {
            1977: ((JUN, 7, "Silver Jubilee of Elizabeth II"),),
            1981: ((JUL, 29, "Wedding of Charles and Diana"),),
            1999: ((DEC, 31, "Millennium Celebrations"),),
            2002: ((JUN, 3, "Golden Jubilee of Elizabeth II"),),
            2011: ((APR, 29, "Wedding of William and Catherine"),),
            2012: ((JUN, 5, "Diamond Jubilee of Elizabeth II"),),
            2022: (
                (JUN, 3, "Platinum Jubilee of Elizabeth II"),
                (SEP, 19, "State Funeral of Queen Elizabeth II"),
            ),
        }

        def _populate(self, year):
            super()._populate(year)

            ...

    For more complex logic, like 4th Monday of January, you can inherit the
    :class:`HolidayBase` class and define your own :meth:`_populate` method.
    See documentation for examples.
    """

    country: str
    """The country's ISO 3166-1 alpha-2 code."""
    market: str
    """The market's ISO 3166-1 alpha-2 code."""
    subdivisions: List[str] = []
    """The subdivisions supported for this country (see documentation)."""
    years: Set[int]
    """The years calculated."""
    expand: bool
    """Whether the entire year is calculated when one date from that year
    is requested."""
    observed: bool
    """Whether dates when public holiday are observed are included."""
    subdiv: Optional[str] = None
    """The subdiv requested."""
    special_holidays: Dict[int, Tuple[Tuple[int, int, str], ...]] = {}
    """A list of the country-wide special (as opposite to regular) holidays for
    a specific year."""
    _deprecated_subdivisions: List[str] = []
    """Other subdivisions whose names are deprecated or aliases of the official
    ones."""
    weekend: Set[int] = {SAT, SUN}
    """Country weekend days."""

    def __init__(
        self,
        years: Optional[Union[int, Iterable[int]]] = None,
        expand: bool = True,
        observed: bool = True,
        subdiv: Optional[str] = None,
        prov: Optional[str] = None,  # deprecated
        state: Optional[str] = None,  # deprecated
    ) -> None:
        """
        :param years:
            The year(s) to pre-calculate public holidays for at instantiation.

        :param subdiv:
            The subdivision (e.g. state or province); not implemented for all
            countries (see documentation).

        :param prov:
            *deprecated* use subdiv instead.

        :param state:
            *deprecated* use subdiv instead.

        :param expand:
            Whether the entire year is calculated when one date from that year
            is requested.

        :param observed:
            Whether to include the dates when public holiday are observed
            (e.g. a holiday falling on a Sunday being observed the
            following Monday). This doesn't work for all countries.

        :return:
            A :class:`HolidayBase` object matching the **country**.
        """
        super().__init__()
        self.observed = observed
        self.expand = expand
        self.subdiv = subdiv or prov or state
        if prov or state:
            warnings.warn(
                "Arguments prov and state are deprecated, use subdiv="
                f"'{prov or state}' instead.",
                DeprecationWarning,
            )
        if not isinstance(self, HolidaySum):
            if (
                subdiv
                and subdiv
                not in self.subdivisions + self._deprecated_subdivisions
            ):
                if hasattr(self, "market"):
                    raise NotImplementedError(
                        f"Market {self.market} does not have subdivision "
                        f"'{subdiv}'"
                    )
                else:
                    raise NotImplementedError(
                        f"Country {self.country} does not have subdivision "
                        f"'{subdiv}'"
                    )
        if isinstance(years, int):
            self.years = {years}
        else:
            self.years = set(years) if years is not None else set()
        for year in self.years.copy():
            self._populate(year)

    def __setattr__(self, key: str, value: Any) -> None:
        if key == "observed" and len(self) > 0:
            dict.__setattr__(self, key, value)
            if value is True:
                # Add (Observed) dates
                years = list(self.years)
                self.years = set()
                self.clear()
                for year in years:
                    self._populate(year)
            else:
                # Remove (Observed) dates
                for k, v in list(self.items()):
                    if v.find("Observed") >= 0:
                        del self[k]
        else:
            return dict.__setattr__(self, key, value)

    def __keytransform__(self, key: DateLike) -> date:
        """Transforms the date from one of the following types:

        * :class:`datetime.date`,
        * :class:`datetime.datetime`,
        * a :class:`str` of any format recognized by
          :func:`dateutil.parser.parse`,
        * or a :class:`float` or :class:`int` representing a POSIX timestamp

        to :class:`datetime.date`, which is how it's stored by the class."""
        if isinstance(key, datetime):
            out_key = key.date()
        elif isinstance(key, date):
            out_key = key
        elif isinstance(key, int) or isinstance(key, float):
            out_key = datetime.utcfromtimestamp(key).date()
        elif isinstance(key, str):
            try:
                out_key = parse(key).date()
            except (ValueError, OverflowError):
                raise ValueError("Cannot parse date from string '%s'" % key)
        else:
            raise TypeError("Cannot convert type '%s' to date." % type(key))

        if self.expand and out_key.year not in self.years:
            self.years.add(out_key.year)
            self._populate(out_key.year)
        return out_key

    def __contains__(self, key: object) -> bool:
        """Return true if date is in self, false otherwise. Accepts a date in
        the following types:

        * :class:`datetime.date`,
        * :class:`datetime.datetime`,
        * a :class:`str` of any format recognized by
          :func:`dateutil.parser.parse`,
        * or a :class:`float` or :class:`int` representing a POSIX timestamp.
        """

        if not isinstance(key, (date, datetime, str, float, int)):
            raise TypeError("Cannot convert type '%s' to date." % type(key))

        return dict.__contains__(
            cast("Mapping[Any, Any]", self), self.__keytransform__(key)
        )

    def __getitem__(self, key: DateLike) -> Any:
        if isinstance(key, slice):
            if not key.start or not key.stop:
                raise ValueError("Both start and stop must be given.")

            start = self.__keytransform__(key.start)
            stop = self.__keytransform__(key.stop)

            if key.step is None:
                step = 1
            elif isinstance(key.step, timedelta):
                step = key.step.days
            elif isinstance(key.step, int):
                step = key.step
            else:
                raise TypeError(
                    "Cannot convert type '%s' to int." % type(key.step)
                )

            if step == 0:
                raise ValueError("Step value must not be zero.")

            date_diff = stop - start
            if date_diff.days < 0 <= step or date_diff.days >= 0 > step:
                step *= -1

            days_in_range = []
            for delta_days in range(0, date_diff.days, step):
                day = start + timedelta(days=delta_days)
                try:
                    self.__getitem__(day)
                    days_in_range.append(day)
                except KeyError:
                    pass
            return days_in_range
        return dict.__getitem__(self, self.__keytransform__(key))

    def __setitem__(self, key: DateLike, value: str) -> None:
        if key in self:
            # If there are multiple holidays on the same date
            # order their names alphabetically.
            delimiter = ", "
            holiday_names = set(self.get(key).split(delimiter))
            holiday_names.add(value)
            value = delimiter.join(sorted(holiday_names))

        dict.__setitem__(self, self.__keytransform__(key), value)

    def update(  # type: ignore[override]
        self, *args: Union[Dict[DateLike, str], List[DateLike], DateLike]
    ) -> None:
        # TODO: fix arguments; should not be *args (cannot properly Type hint)
        """Update the object, overwriting existing dates.

        :param:
            Either another dictionary object where keys are dates and values
            are holiday names, or a single date (or a list of dates) for which
            the value will be set to "Holiday".

            Dates can be expressed in one or more of the following types:

            * :class:`datetime.date`,
            * :class:`datetime.datetime`,
            * a :class:`str` of any format recognized by
              :func:`dateutil.parser.parse`,
            * or a :class:`float` or :class:`int` representing a POSIX
              timestamp.
        """
        args_list = list(args)
        for arg in args_list:
            if isinstance(arg, dict):
                for key, value in list(arg.items()):
                    self[key] = value
            elif isinstance(arg, list):
                for item in arg:
                    self[item] = "Holiday"
            else:
                self[arg] = "Holiday"

    def append(
        self, *args: Union[Dict[DateLike, str], List[DateLike], DateLike]
    ) -> None:
        """Alias for :meth:`update` to mimic list type."""
        return self.update(*args)

    def get(
        self,
        key: DateLike,
        default: Union[str, Any] = None,
    ) -> Union[str, Any]:
        """Return the holiday name for a date if date is a holiday, else
        default. If default is not given, it defaults to None, so that this
        method never raises a KeyError. If more than one holiday is present,
        they are separated by a comma.

        :param key:
            The date expressed in one of the following types:

            * :class:`datetime.date`,
            * :class:`datetime.datetime`,
            * a :class:`str` of any format recognized by
              :func:`dateutil.parser.parse`,
            * or a :class:`float` or :class:`int` representing a POSIX
              timestamp.

        :param default:
            The default value to return if no value is found.
        """
        return dict.get(
            self,
            self.__keytransform__(key),
            default,
        )

    def get_list(self, key: DateLike) -> List[str]:
        """Return a list of all holiday names for a date if date is a holiday,
        else empty string.

        :param key:
            The date expressed in one of the following types:

            * :class:`datetime.date`,
            * :class:`datetime.datetime`,
            * a :class:`str` of any format recognized by
              :func:`dateutil.parser.parse`,
            * or a :class:`float` or :class:`int` representing a POSIX
              timestamp.
        """
        return [h for h in self.get(key, "").split(", ") if h]

    def get_named(self, name: str) -> List[date]:
        """Return a list of all holiday dates matching the provided holiday
        name. The match will be made case insensitively and partial matches
        will be included.

        :param name:
            The holiday's name to try to match.

        :return:
            A list of all holiday dates matching the provided holiday name.
        """
        original_expand = self.expand
        self.expand = False
        matches = [
            key for key in self if name.lower() in str(self[key]).lower()
        ]
        self.expand = original_expand
        return matches

    def pop(
        self,
        key: DateLike,
        default: Union[str, Any] = None,
    ) -> Union[str, Any]:
        """If date is a holiday, remove it and return its date, else return
        default.

        :param key:
            The date expressed in one of the following types:

            * :class:`datetime.date`,
            * :class:`datetime.datetime`,
            * a :class:`str` of any format recognized by
              :func:`dateutil.parser.parse`,
            * or a :class:`float` or :class:`int` representing a POSIX
              timestamp.

        :param default:
            The default value to return if no match is found.

        :return:
            The date removed.

        :raise:
            KeyError if date is not a holiday and default is not given.
        """
        if default is None:
            return dict.pop(self, self.__keytransform__(key))
        return dict.pop(self, self.__keytransform__(key), default)

    def pop_named(self, name: str) -> List[date]:
        """Remove (no longer treat at as holiday) all dates matching the
        provided holiday name. The match will be made case insensitively and
        partial matches will be removed.

        :param name:
            The holiday's name to try to match.

        :param default:
            The default value to return if no match is found.

        :return:
            A list of dates removed.

        :raise:
            KeyError if date is not a holiday and default is not given.
        """
        to_pop = self.get_named(name)
        if not to_pop:
            raise KeyError(name)
        for key in to_pop:
            self.pop(key)
        return to_pop

    def __eq__(self, other: object) -> bool:
        return dict.__eq__(self, other) and self.__dict__ == other.__dict__

    def __ne__(self, other: object) -> bool:
        return dict.__ne__(self, other) or self.__dict__ != other.__dict__

    def __add__(
        self, other: Union[int, "HolidayBase", "HolidaySum"]
    ) -> "HolidayBase":
        """Add another dictionary of public holidays creating a
        :class:`HolidaySum` object.

        :param other:
            The dictionary of public holiday to be added.

        :return:
            A :class:`HolidaySum` object unless the other object cannot be
            added, then :class:`self`.
        """
        if isinstance(other, int) and other == 0:
            # Required to sum() list of holidays
            # sum([h1, h2]) is equivalent to (0 + h1 + h2)
            return self
        elif not isinstance(other, (HolidayBase, HolidaySum)):
            raise TypeError(
                "Holiday objects can only be added with other Holiday objects"
            )
        return HolidaySum(self, other)

    def __radd__(self, other: Any) -> "HolidayBase":
        return self.__add__(other)

    def _populate(self, year: int) -> None:
        """This is a private class that populates (generates and adds) holidays
        for a given year. To keep things fast, it assumes that no holidays for
        the year have already been populated. It is required to be called
        internally by any country populate() method, while should not be called
        directly from outside.
        To add holidays to an object, use the update() method:

        :param year:
            The year to populate with holidays.

        >>> from holidays import country_holidays
        >>> us_holidays = country_holidays('US', years=2020)
        # to add new holidays to the object:
        >>> us_holidays.update(country_holidays('US', years=2021))
        """

        # Special holidays list.
        for month, day, name in self.special_holidays.get(year, ()):
            self[date(year, month, day)] = name

    def _is_weekend(self, *args):
        """
        Returns True if date's week day is a weekend day.
        Returns False otherwise.
        """
        dt = args[0] if len(args) == 1 else date(*args)

        return dt.weekday() in self.weekend

    def __reduce__(self) -> Union[str, Tuple[Any, ...]]:
        return super().__reduce__()

    def __repr__(self) -> str:
        _repr = ""
        if len(self) == 0:
            if hasattr(self, "market"):
                _repr = f"holidays.financial_holidays({self.market!r}"
                if self.subdiv:
                    _repr += f", subdiv={self.subdiv!r}"
                _repr += ")"
            if hasattr(self, "country"):
                if _repr:
                    _repr += " + "
                _repr += f"holidays.country_holidays({self.country!r}"
                if self.subdiv:
                    _repr += f", subdiv={self.subdiv!r}"
                _repr += ")"
            return _repr
        return super().__repr__()

    def __str__(self) -> str:
        if len(self) == 0:
            return str(self.__dict__)
        return super().__str__()


class HolidaySum(HolidayBase):
    """
    Returns a :class:`dict`-like object resulting from the addition of two or
    more individual dictionaries of public holidays. The original dictionaries
    are available as a :class:`list` in the attribute :attr:`holidays,` and
    :attr:`country` and :attr:`subdiv` attributes are added
    together and could become :class:`list` s. Holiday names, when different,
    are merged. All years are calculated (expanded) for all operands.
    """

    country: Union[str, List[str]]  # type: ignore[assignment]
    """Countries included in the addition."""
    market: Union[str, List[str]]  # type: ignore[assignment]
    """Markets included in the addition."""
    subdiv: Optional[Union[str, List[str]]]  # type: ignore[assignment]
    """Subdivisions included in the addition."""
    holidays: List[HolidayBase]
    """The original HolidayBase objects included in the addition."""
    years: Set[int]
    """The years calculated."""

    def __init__(
        self,
        h1: Union[HolidayBase, "HolidaySum"],
        h2: Union[HolidayBase, "HolidaySum"],
    ) -> None:
        """
        :param h1:
            The first HolidayBase object to add.

        :param h2:
            The other HolidayBase object to add.

        Example:

        >>> from holidays import country_holidays
        >>> nafta_holidays = country_holidays('US', years=2020) + \
country_holidays('CA') + country_holidays('MX')
        >>> dates = sorted(nafta_holidays.items(), key=lambda x: x[0])
        >>> from pprint import pprint
        >>> pprint(dates[:10], width=72)
        [(datetime.date(2020, 1, 1), "Año Nuevo [New Year's Day]"),
         (datetime.date(2020, 1, 20), 'Martin Luther King Jr. Day'),
         (datetime.date(2020, 2, 3),
          'Día de la Constitución [Constitution Day] (Observed)'),
         (datetime.date(2020, 2, 5),
          'Día de la Constitución [Constitution Day]'),
         (datetime.date(2020, 2, 17), "Washington's Birthday, Family Day"),
         (datetime.date(2020, 3, 16),
          "Natalicio de Benito Juárez [Benito Juárez's birthday] (Observed)"),
         (datetime.date(2020, 3, 21),
          "Natalicio de Benito Juárez [Benito Juárez's birthday]"),
         (datetime.date(2020, 4, 10), 'Good Friday'),
         (datetime.date(2020, 5, 1), 'Día del Trabajo [Labour Day]'),
         (datetime.date(2020, 5, 18), 'Victoria Day')]
        """
        # store originals in the holidays attribute
        self.holidays = []
        for operand in (h1, h2):
            if isinstance(operand, HolidaySum):
                for h in operand.holidays:
                    self.holidays.append(h)
            else:
                self.holidays.append(operand)
        kwargs: Dict[str, Any] = {}
        # join years, expand and observed
        kwargs["years"] = h1.years | h2.years
        kwargs["expand"] = h1.expand or h2.expand
        kwargs["observed"] = h1.observed or h2.observed
        # join country and subdivisions data
        # TODO this way makes no sense: joining Italy Catania (IT, CA) with
        # USA Mississippi (US, MS) and USA Michigan (US, MI) yields
        # country=["IT", "US"] and subdiv=["CA", "MS", "MI"], which could very
        # well be California and Messina and Milano, or Catania, Mississippi
        # and Milano, or ... you get the picture.
        # Same goes when countries and markets are being mixed (working, yet
        # still nonsensical)
        for attr in ("country", "market", "subdiv"):
            if (
                getattr(h1, attr, None)
                and getattr(h2, attr, None)
                and getattr(h1, attr) != getattr(h2, attr)
            ):
                a1 = (
                    getattr(h1, attr)
                    if isinstance(getattr(h1, attr), list)
                    else [getattr(h1, attr)]
                )
                a2 = (
                    getattr(h2, attr)
                    if isinstance(getattr(h2, attr), list)
                    else [getattr(h2, attr)]
                )
                kwargs[attr] = a1 + a2
            else:
                arg = getattr(h1, attr, None) or getattr(h2, attr, None)
                if arg:
                    kwargs[attr] = arg

        if "country" in kwargs:
            self.country = kwargs.pop("country")
        if "market" in kwargs:
            self.market = kwargs.pop("market")

        HolidayBase.__init__(self, **kwargs)

    def _populate(self, year):
        for h in self.holidays[::-1]:
            h._populate(year)
            self.update(cast("Dict[DateLike, str]", h))



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import MO
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, FEB, MAY, JUN, AUG, SEP, NOV, DEC, TUE
from holidays.constants import WED, THU, FRI
from holidays.holiday_base import HolidayBase


class DominicanRepublic(HolidayBase):
    """
    http://ojd.org.do/Normativas/LABORAL/Leyes/Ley%20No.%20%20139-97.pdf
    https://es.wikipedia.org/wiki/Rep%C3%BAblica_Dominicana#D%C3%ADas_festivos_nacionales
    """

    country = "DO"

    @staticmethod
    def __change_day_by_law(holiday, latest_days=(THU, FRI)):
        # Law No. 139-97 - Holidays Dominican Republic - Jun 27, 1997
        if holiday >= date(1997, 6, 27):
            if holiday.weekday() in {TUE, WED}:
                holiday -= rd(weekday=MO(-1))
            elif holiday.weekday() in latest_days:
                holiday += rd(weekday=MO(1))
        return holiday

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day
        self[date(year, JAN, 1)] = "Año Nuevo [New Year's Day]"

        # Epiphany
        epiphany_day = self.__change_day_by_law(date(year, JAN, 6))
        self[epiphany_day] = "Día de los Santos Reyes [Epiphany]"

        # Lady of Altagracia
        self[date(year, JAN, 21)] = "Día de la Altagracia [Lady of Altagracia]"

        # Juan Pablo Duarte Day
        duarte_day = self.__change_day_by_law(date(year, JAN, 26))
        self[duarte_day] = "Día de Duarte [Juan Pablo Duarte Day]"

        # Independence Day
        self[date(year, FEB, 27)] = "Día de Independencia [Independence Day]"

        # Good Friday
        self[easter(year) + rd(days=-2)] = "Viernes Santo [Good Friday]"

        # Labor Day
        labor_day = self.__change_day_by_law(date(year, MAY, 1), (3, 4, 6))
        self[labor_day] = "Día del Trabajo [Labor Day]"

        # Feast of Corpus Christi
        self[date(year, JUN, 11)] = "Corpus Christi [Feast of Corpus Christi]"

        # Restoration Day
        # Judgment No. 14 of Feb 20, 2008 of the Supreme Court of Justice
        restoration_day = (
            date(year, AUG, 16)
            if ((year - 2000) % 4 == 0) and year < 2008
            else self.__change_day_by_law(date(year, AUG, 16))
        )
        self[restoration_day] = "Día de la Restauración [Restoration Day]"

        # Our Lady of Mercedes Day
        self[
            date(year, SEP, 24)
        ] = "Día de las Mercedes [Our Lady of Mercedes Day]"

        # Constitution Day
        constitution_day = self.__change_day_by_law(date(year, NOV, 6))
        self[constitution_day] = "Día de la Constitución [Constitution Day]"

        # Christmas Day
        self[date(year, DEC, 25)] = "Día de Navidad [Christmas Day]"


class DO(DominicanRepublic):
    pass


class DOM(DominicanRepublic):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)
#  Copyright: Kateryna Golovanova <kate@kgthreads.com>, 2022

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import AUG, DEC, FEB, JAN, MAR, MAY, NOV, SEP
from holidays.holiday_base import HolidayBase


class Liechtenstein(HolidayBase):
    """
    Liechtenstein holidays.
    See https://en.wikipedia.org/wiki/Public_holidays_in_Liechtenstein
    for details.
    """

    country = "LI"

    def _populate(self, year):
        # New Year's Day.
        self[date(year, JAN, 1)] = "Neujahr"

        # Saint Berchtold's Day.
        self[date(year, JAN, 2)] = "Berchtoldstag"

        # Epiphany.
        self[date(year, JAN, 6)] = "Drei Könige"

        # Candlemas.
        self[date(year, FEB, 2)] = "Mariä Lichtmess"

        easter_date = easter(year)
        # Shrove Tuesday.
        self[easter_date + rd(days=-47)] = "Fasnachtsdienstag"

        # Saint Joseph's Day.
        self[date(year, MAR, 19)] = "Josefstag"

        # Good Friday.
        self[easter_date + rd(days=-2)] = "Karfreitag"

        # Easter.
        self[easter_date] = "Ostersonntag"

        # Easter Monday.
        self[easter_date + rd(days=+1)] = "Ostermontag"

        # Labor Day.
        self[date(year, MAY, 1)] = "Tag der Arbeit"

        # Ascension Day.
        self[easter_date + rd(days=+39)] = "Auffahrt"

        # Pentecost.
        self[easter_date + rd(days=+49)] = "Pfingstsonntag"

        # Whit Monday.
        self[easter_date + rd(days=+50)] = "Pfingstmontag"

        # Corpus Christi.
        self[easter_date + rd(days=+60)] = "Fronleichnam"

        # National Day.
        self[date(year, AUG, 15)] = "Staatsfeiertag"

        # Nativity of Mary.
        self[date(year, SEP, 8)] = "Maria Geburt"

        # All Saints Day.
        self[date(year, NOV, 1)] = "Allerheiligen"

        # Feast of the Immaculate Conception.
        self[date(year, DEC, 8)] = "Maria Empfängnis"

        # Christmas Eve.
        self[date(year, DEC, 24)] = "Heiliger Abend"

        # Christmas Day.
        self[date(year, DEC, 25)] = "Weihnachten"

        # St. Stephen's Day.
        self[date(year, DEC, 26)] = "Stefanstag"

        # New Year's Eve.
        self[date(year, DEC, 31)] = "Silvester"


class LI(Liechtenstein):
    pass


class LIE(Liechtenstein):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import MO, FR
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT
from holidays.constants import NOV, DEC, SUN
from holidays.holiday_base import HolidayBase


class SouthAfrica(HolidayBase):
    """
    http://www.gov.za/about-sa/public-holidays
    https://en.wikipedia.org/wiki/Public_holidays_in_South_Africa
    """

    country = "ZA"
    special_holidays = {
        1999: (
            (JUN, 2, "National and provincial government elections"),
            (DEC, 31, "Y2K changeover"),
        ),
        2000: ((JAN, 2, "Y2K changeover"),),
        2004: ((APR, 14, "National and provincial government elections"),),
        2006: ((MAR, 1, "Local government elections"),),
        2008: ((MAY, 2, "Public holiday by presidential decree"),),
        2009: ((APR, 22, "National and provincial government elections"),),
        2011: (
            (MAY, 18, "Local government elections"),
            (DEC, 27, "Public holiday by presidential decree"),
        ),
        2014: ((MAY, 7, "National and provincial government elections"),),
        2016: (
            (AUG, 3, "Local government elections"),
            (DEC, 27, "Public holiday by presidential decree"),
        ),
        2019: ((MAY, 8, "National and provincial government elections"),),
        2021: ((NOV, 1, "Municipal elections"),),
        2022: ((DEC, 27, "Public holiday by presidential decree"),),
    }

    def _populate(self, year):
        # Observed since 1910, with a few name changes
        if year <= 1909:
            return
        super()._populate(year)

        self[date(year, JAN, 1)] = "New Year's Day"

        easter_date = easter(year)
        self[easter_date + rd(days=-2)] = "Good Friday"
        if year >= 1980:
            name = "Family Day"
        else:
            name = "Easter Monday"
        self[easter_date + rd(days=+1)] = name

        if year <= 1951:
            name = "Dingaan's Day"
        elif year <= 1979:
            name = "Day of the Covenant"
        elif year <= 1994:
            name = "Day of the Vow"
        else:
            name = "Day of Reconciliation"
        self[date(year, DEC, 16)] = name

        self[date(year, DEC, 25)] = "Christmas Day"

        if year >= 1980:
            name = "Day of Goodwill"
        else:
            name = "Boxing Day"
        self[date(year, DEC, 26)] = name

        # Observed since 1995/1/1
        if year >= 1995:
            self[date(year, MAR, 21)] = "Human Rights Day"
            self[date(year, APR, 27)] = "Freedom Day"
            self[date(year, MAY, 1)] = "Workers' Day"
            self[date(year, JUN, 16)] = "Youth Day"
            self[date(year, AUG, 9)] = "National Women's Day"
            self[date(year, SEP, 24)] = "Heritage Day"

        # As of 1995/1/1, whenever a public holiday falls on a Sunday,
        # it rolls over to the following Monday
        if self.observed and year >= 1995:
            for k, v in list(self.items()):
                if k.weekday() != SUN or k.year != year:
                    continue
                dt = k + rd(days=+1)
                if dt in self:
                    continue
                self[dt] = v + " (Observed)"

        # Historic public holidays no longer observed
        if 1952 <= year <= 1973:
            self[date(year, APR, 6)] = "Van Riebeeck's Day"
        elif 1980 <= year <= 1994:
            self[date(year, APR, 6)] = "Founder's Day"

        if 1987 <= year <= 1989:
            # observed on first Friday in May
            self[(date(year, MAY, 1) + rd(weekday=FR))] = "Workers' Day"

        if year <= 1993:
            self[easter_date + rd(days=+40)] = "Ascension Day"

        if year <= 1951:
            self[date(year, MAY, 24)] = "Empire Day"

        if year <= 1960:
            self[date(year, MAY, 31)] = "Union Day"
        elif year <= 1993:
            self[date(year, MAY, 31)] = "Republic Day"

        if 1952 <= year <= 1960:
            # observed on second Monday in July
            self[
                (date(year, JUL, 1) + rd(weekday=MO(+2)))
            ] = "Queen's Birthday"

        if 1961 <= year <= 1973:
            self[date(year, JUL, 10)] = "Family Day"

        if year <= 1951:
            # observed on first Monday in August
            self[(date(year, AUG, 1) + rd(weekday=MO))] = "King's Birthday"

        if 1952 <= year <= 1979:
            # observed on first Monday in September
            self[(date(year, SEP, 1) + rd(weekday=MO))] = "Settlers' Day"

        if 1952 <= year <= 1993:
            self[date(year, OCT, 10)] = "Kruger Day"


class ZA(SouthAfrica):
    pass


class ZAF(SouthAfrica):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, DEC
from holidays.holiday_base import HolidayBase


class Denmark(HolidayBase):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Denmark
    """

    country = "DK"

    def _populate(self, year):
        super()._populate(year)

        easter_date = easter(year)
        # Public holidays
        self[date(year, JAN, 1)] = "Nytårsdag"
        self[easter_date + rd(days=-7)] = "Palmesøndag"
        self[easter_date + rd(days=-3)] = "Skærtorsdag"
        self[easter_date + rd(days=-2)] = "Langfredag"
        self[easter_date] = "Påskedag"
        self[easter_date + rd(days=+1)] = "Anden påskedag"
        self[easter_date + rd(days=+26)] = "Store bededag"
        self[easter_date + rd(days=+39)] = "Kristi himmelfartsdag"
        self[easter_date + rd(days=+49)] = "Pinsedag"
        self[easter_date + rd(days=+50)] = "Anden pinsedag"
        self[date(year, DEC, 25)] = "Juledag"
        self[date(year, DEC, 26)] = "Anden juledag"


class DK(Denmark):
    pass


class DNK(Denmark):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)
#  Copyright: Kateryna Golovanova <kate@kgthreads.com>, 2022

from datetime import date

from dateutil.easter import EASTER_ORTHODOX, easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import DEC, JAN, JUN, MAR, MAY, NOV, SUN
from holidays.holiday_base import HolidayBase
from holidays.utils import _islamic_to_gre


class BosniaAndHerzegovina(HolidayBase):
    """
    Bosnia and Herzegovina holidays.
    See https://en.wikipedia.org/wiki/Public_holidays_in_Bosnia_and_Herzegovina
    for details.
    """

    country = "BA"
    subdivisions = [
        "BD",
        "FBiH",
        "RS",
    ]

    def _populate(self, year):

        # New Year's Day.
        self[date(year, JAN, 1)] = "Nova Godina"
        self[date(year, JAN, 2)] = "Drugi dan Nove Godine"

        if self.observed and date(year, JAN, 1).weekday() == SUN:
            self[date(year, JAN, 1) + rd(days=+2)] = "Treći dan Nove Godine"

        # Labor Day.
        may_1 = date(year, MAY, 1)
        self[may_1] = "Dan rada"
        self[may_1 + rd(days=+1)] = "Drugi dan Dana rada"

        if self.observed and may_1.weekday() == SUN:
            self[may_1 + rd(days=+2)] = "Treći dan Dana rada"

        if self.subdiv == "FBiH":
            # Independence Day.
            self[date(year, MAR, 1)] = "Dan nezavisnosti"

            easter_date = easter(year)
            # Catholic Good Friday.
            self[easter_date + rd(days=-2)] = "Veliki Petak (Katolički)"

            # Catholic Easter.
            self[easter_date] = "Uskrs (Katolički)"
            self[easter_date + rd(days=+1)] = "Uskrsni ponedjeljak (Katolički)"

            # Corpus Cristi.
            self[
                easter_date + rd(days=+60)
            ] = "Tijelovo (Tijelo i Krv Kristova)"

            # Eid al-Fitr.
            # Date of observance is announced yearly, this is an estimate.
            for dt in _islamic_to_gre(year, 10, 1):
                self[dt] = "Ramazanski Bajram"
                self[dt + rd(days=+1)] = "Drugi Dan Ramazanski Bajram"

            # Eid ul-Adha.
            # Date of observance is announced yearly, this is an estimate.
            name = "Kurban Bajram"
            for dt in _islamic_to_gre(year, 12, 10):
                self[dt] = name
                for d in range(1, 4):
                    self[dt + rd(days=+d)] = name

            # Islamic New Year.
            for dt in _islamic_to_gre(year, 1, 1):
                self[dt] = "Muslimanska Nova Godina"

            # All Saints Day.
            self[date(year, 11, 1)] = "Svi Sveti"

            # All Souls Day.
            self[date(year, 11, 2)] = "Dušni dan"

            # Statehood Day.
            self[date(year, NOV, 25)] = "Dan državnosti"

            # Catholic Christmas.
            self[date(year, DEC, 25)] = "Božić (Katolički)"

            # St. Stephen's Day.
            self[date(year, DEC, 26)] = "Stipandan (Stjepandan)"

        elif self.subdiv == "RS":
            # Orthodox Christmas Eve.
            self[date(year, JAN, 6)] = "Pravoslavno Badnje veče"

            # Orthodox Christmas.
            self[date(year, JAN, 7)] = "Božić (Божић)"

            # Republic day.
            self[date(year, JAN, 9)] = "Dan Republike"

            # Orthodox New Year.
            self[date(year, JAN, 14)] = "Pravoslavna Nova Godina"

            easter_date = easter(year, method=EASTER_ORTHODOX)
            # Orthodox Good Friday.
            self[easter_date + rd(days=-2)] = "Veliki Petak (Pravoslavni)"

            # Orthodox Easter.
            self[easter_date] = "Vaskrs (Pravoslavni)"
            self[
                easter_date + rd(days=+1)
            ] = "Uskrsni ponedjeljak (Pravoslavni)"

            # Victory Day.
            self[date(year, MAY, 9)] = "Dan pobjede"

            # St. Vitus Day.
            self[date(year, JUN, 28)] = "Vidovdan"

            # Dayton Agreement Day.
            self[date(year, NOV, 21)] = (
                "Dan uspostave Opšteg okvirnog sporazuma za mir u "
                "Bosni i Hercegovini"
            )


class BA(BosniaAndHerzegovina):
    pass


class BIH(BosniaAndHerzegovina):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#           tasnim<tasnimislam1999@gmail.com>
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from holidays.constants import FEB, MAR, APR, MAY, AUG, DEC
from holidays.holiday_base import HolidayBase


class Bangladesh(HolidayBase):
    """
    https://mopa.gov.bd/sites/default/files/files/mopa.gov.bd/public_holiday/61c35b73_e335_462a_9bcf_4695b23b6d82/reg4-2019-212.PDF
    https://en.wikipedia.org/wiki/Public_holidays_in_Bangladesh
    """

    country = "BD"

    def _populate(self, year):
        super()._populate(year)

        # 21st Feb
        self[date(year, FEB, 21)] = "International Mother's language Day"

        # 17th March
        self[date(year, MAR, 17)] = (
            "Sheikh Mujibur Rahman's Birthday " "and Children's Day"
        )

        # 26th March
        self[date(year, MAR, 26)] = "Independence Day"

        # 14th April
        self[date(year, APR, 14)] = "Bengali New Year's Day"

        # 1st May
        self[date(year, MAY, 1)] = "May Day"

        # 15th AUG
        self[date(year, AUG, 15)] = "National Mourning Day"

        # 16th Dec
        self[date(year, DEC, 16)] = "Victory Day"


class BD(Bangladesh):
    pass


class BGD(Bangladesh):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date, datetime

from dateutil import rrule
from dateutil.easter import easter
from dateutil.relativedelta import FR, SA, SU
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAR, MAY, JUN, OCT, DEC
from holidays.holiday_base import HolidayBase


class Sweden(HolidayBase):
    """
    Swedish holidays.
    Note that holidays falling on a sunday are "lost",
    it will not be moved to another day to make up for the collision.
    In Sweden, ALL sundays are considered a holiday
    (https://sv.wikipedia.org/wiki/Helgdagar_i_Sverige).
    Initialize this class with include_sundays=False
    to not include sundays as a holiday.
    Primary sources:
    https://sv.wikipedia.org/wiki/Helgdagar_i_Sverige and
    http://www.riksdagen.se/sv/dokument-lagar/dokument/svensk-forfattningssamling/lag-1989253-om-allmanna-helgdagar_sfs-1989-253
    """

    country = "SE"

    def __init__(self, include_sundays=True, **kwargs):
        """
        :param include_sundays: Whether to consider sundays as a holiday
        (which they are in Sweden)
        :param kwargs:
        """
        self.include_sundays = include_sundays
        HolidayBase.__init__(self, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        if self.include_sundays:  # Optionally add all Sundays of the year.
            year_first_day = datetime(year, JAN, 1)
            year_last_day = datetime(year, DEC, 31)

            # Get all Sundays including first/last day of the year cases.
            sundays = rrule.rrule(
                rrule.WEEKLY, byweekday=SU, dtstart=year_first_day
            ).between(year_first_day, year_last_day, inc=True)
            for sunday in sundays:
                self[sunday.date()] = "Söndag"

        # ========= Static holidays =========
        self[date(year, JAN, 1)] = "Nyårsdagen"

        self[date(year, JAN, 6)] = "Trettondedag jul"

        # Source: https://sv.wikipedia.org/wiki/F%C3%B6rsta_maj
        if year >= 1939:
            self[date(year, MAY, 1)] = "Första maj"

        # Source: https://sv.wikipedia.org/wiki/Sveriges_nationaldag
        if year >= 2005:
            self[date(year, JUN, 6)] = "Sveriges nationaldag"

        self[date(year, DEC, 24)] = "Julafton"
        self[date(year, DEC, 25)] = "Juldagen"
        self[date(year, DEC, 26)] = "Annandag jul"
        self[date(year, DEC, 31)] = "Nyårsafton"

        # ========= Moving holidays =========
        easter_date = easter(year)
        self[easter_date + rd(days=-2)] = "Långfredagen"
        self[easter_date] = "Påskdagen"
        self[easter_date + rd(days=+1)] = "Annandag påsk"
        self[easter_date + rd(days=+39)] = "Kristi himmelsfärdsdag"
        self[easter_date + rd(days=+49)] = "Pingstdagen"
        if year <= 2004:
            self[easter_date + rd(days=+50)] = "Annandag pingst"

        # Source:
        # https://sv.wikipedia.org/wiki/Midsommarafton
        # https://www.nordiskamuseet.se/aretsdagar/midsommarafton
        if year >= 1953:
            # Midsummer evening. Friday between June 19th and June 25th
            self[date(year, JUN, 19) + rd(weekday=FR)] = "Midsommarafton"
            # Midsummer day. Saturday between June 20th and June 26th
            self[date(year, JUN, 20) + rd(weekday=SA)] = "Midsommardagen"
        else:
            self[date(year, JUN, 23)] = "Midsommarafton"
            self[date(year, JUN, 24)] = "Midsommardagen"

        # All saints day. Saturday between October 31th and November 6th
        self[date(year, OCT, 31) + rd(weekday=SA)] = "Alla helgons dag"

        if year <= 1953:
            self[date(year, MAR, 25)] = "Jungfru Marie bebådelsedag"


class SE(Sweden):
    pass


class SWE(Sweden):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import FR, SA
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAY, JUN, OCT, DEC
from holidays.holiday_base import HolidayBase


class Finland(HolidayBase):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Finland
    """

    country = "FI"

    def _populate(self, year):
        super()._populate(year)

        easter_date = easter(year)

        self[date(year, JAN, 1)] = "Uudenvuodenpäivä"
        self[date(year, JAN, 6)] = "Loppiainen"
        self[easter_date + rd(days=-2)] = "Pitkäperjantai"
        self[easter_date] = "Pääsiäispäivä"
        self[easter_date + rd(days=+1)] = "2. pääsiäispäivä"
        self[date(year, MAY, 1)] = "Vappu"
        self[easter_date + rd(days=+39)] = "Helatorstai"
        self[easter_date + rd(days=+49)] = "Helluntaipäivä"
        self[date(year, JUN, 20) + rd(weekday=SA)] = "Juhannuspäivä"
        self[date(year, OCT, 31) + rd(weekday=SA)] = "Pyhäinpäivä"
        self[date(year, DEC, 6)] = "Itsenäisyyspäivä"
        self[date(year, DEC, 25)] = "Joulupäivä"
        self[date(year, DEC, 26)] = "Tapaninpäivä"

        # Juhannusaatto (Midsummer Eve) and Jouluaatto (Christmas Eve) are not
        # official holidays, but are de facto.
        self[date(year, JUN, 19) + rd(weekday=FR)] = "Juhannusaatto"
        self[date(year, DEC, 24)] = "Jouluaatto"


class FI(Finland):
    pass


class FIN(Finland):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date, timedelta

from dateutil.easter import easter
from dateutil.relativedelta import MO, TH, FR
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP
from holidays.constants import OCT, NOV, DEC, WED, FRI
from holidays.holiday_base import HolidayBase


class NewYorkStockExchange(HolidayBase):
    # Official regulations:
    # https://www.nyse.com/publicdocs/nyse/regulation/nyse/NYSE_Rules.pdf
    # https://www.nyse.com/markets/hours-calendars
    # Historical data:
    # s3.amazonaws.com/armstrongeconomics-wp/2013/07/NYSE-Closings.pdf

    market = "NYSE"

    def __init__(self, **kwargs):
        HolidayBase.__init__(self, **kwargs)

    def _get_observed(self, d):
        wdnum = d.isoweekday()
        if wdnum == 6:
            return d + rd(weekday=FR(-1))
        if wdnum == 7:
            return d + rd(weekday=MO(+1))
        return d

    def _set_observed_date(self, holiday_date, name):
        date_obs = self._get_observed(holiday_date)
        if date_obs == holiday_date:
            self[holiday_date] = name
        else:
            self[date_obs] = name + " (Observed)"

    def _populate(self, year):
        super()._populate(year)

        ##############################################################
        # REGULAR HOLIDAYS
        ##############################################################

        # NYD
        # This year's New Year Day.
        self._set_observed_date(date(year, JAN, 1), "New Year's Day")

        # https://www.nyse.com/publicdocs/nyse/regulation/nyse/NYSE_Rules.pdf
        # As per Rule 7.2.: check if next year's NYD falls on Saturday and
        # needs to be observed on Friday (Dec 31 of previous year).
        dec_31 = date(year, DEC, 31)
        if dec_31.weekday() == FRI:
            self._set_observed_date(dec_31 + rd(days=+1), "New Year's Day")

        # MLK - observed 1998 - 3rd Monday of Jan
        if year >= 1998:
            self[
                date(year, JAN, 1) + rd(weekday=MO(3))
            ] = "Martin Luther King Jr. Day"

        # LINCOLN BIRTHDAY: observed 1896 - 1953 and 1968, Feb 12 (observed)
        if (1896 <= year <= 1953) or year == 1968:
            lincoln = date(year, FEB, 12)
            self._set_observed_date(lincoln, "Lincoln's Birthday")

        # WASHINGTON'S BIRTHDAY: Feb 22 (obs) until 1971, then 3rd Mon of Feb
        if year < 1971:
            wash = date(year, FEB, 22)
            self._set_observed_date(wash, "Washington's Birthday")
        else:
            self[
                date(year, FEB, 1) + rd(weekday=MO(3))
            ] = "Washington's Birthday"

        # GOOD FRIDAY - closed every year except 1898, 1906, and 1907
        e = easter(year)
        if year not in {1898, 1906, 1907}:
            self[e - rd(days=2)] = "Good Friday"

        # MEM DAY (May 30) - closed every year since 1873
        # last Mon in May since 1971
        if 1873 <= year < 1971:
            memday = date(year, MAY, 30)
            self._set_observed_date(memday, "Memorial Day")
        else:
            self[date(year, MAY, 31) + rd(weekday=MO(-1))] = "Memorial Day"

        # FLAG DAY: June 14th 1916 - 1953
        if 1916 <= year <= 1953:
            flagday = date(year, JUN, 14)
            self._set_observed_date(flagday, "Flag Day")

        # JUNETEENTH: since 2021
        if year >= 2021:
            juneteenth = date(year, JUN, 19)
            self._set_observed_date(
                juneteenth, "Juneteenth National Independence Day"
            )

        # INDEPENDENCE DAY (July 4) - history suggests closed every year
        j4th = date(year, JUL, 4)
        self._set_observed_date(j4th, "Independence Day")

        # LABOR DAY - first mon in Sept, since 1887
        if year >= 1887:
            self[date(year, SEP, 1) + rd(weekday=MO(1))] = "Labor Day"

        # COLUMBUS DAY/INDIGENOUS PPL DAY: Oct 12 - closed 1909-1953
        if 1909 <= year <= 1953:
            colday = date(year, OCT, 12)
            self._set_observed_date(colday, "Columbus Day")

        # ELECTION DAY: Tuesday after first Monday in November (2 U.S. Code §7)
        # closed until 1969, then closed pres years 1972-80
        if year <= 1968 or year in {1972, 1976, 1980}:
            self[
                date(year, NOV, 1) + rd(weekday=MO) + rd(days=+1)
            ] = "Election Day"

        # VETERAN'S DAY: Nov 11 - closed 1918, 1921, 1934-1953
        if year in {1918, 1921} or (1934 <= year <= 1953):
            vetday = date(year, NOV, 11)
            self._set_observed_date(vetday, "Veteran's Day")

        # THXGIVING DAY: 4th Thurs in Nov - closed every year
        self[date(year, NOV, 1) + rd(weekday=TH(4))] = "Thanksgiving Day"

        # XMAS DAY: Dec 25th - every year
        xmas = date(year, DEC, 25)
        self._set_observed_date(xmas, "Christmas Day")

        ##############################################################
        # SPECIAL HOLIDAYS
        ##############################################################
        if year == 1888:
            self[date(year, MAR, 12)] = "Blizzard of 1888"
            self[date(year, MAR, 13)] = "Blizzard of 1888"
            self[date(year, NOV, 30)] = "Thanksgiving Friday 1888"
        elif year == 1889:
            self[date(year, APR, 29)] = "Centennial of Washington Inauguration"
            self[date(year, APR, 30)] = "Centennial of Washington Inauguration"
            self[date(year, MAY, 1)] = "Centennial of Washington Inauguration"
        elif year == 1892:
            self[date(year, OCT, 12)] = "Columbian Celebration"
            self[date(year, OCT, 21)] = "Columbian Celebration"
        elif year == 1893:
            self[date(year, APR, 27)] = "Columbian Celebration"
        elif year == 1897:
            self[date(year, APR, 27)] = "Grant's Birthday"
        elif year == 1898:
            self[date(year, MAY, 4)] = "Charter Day"
        elif year == 1899:
            self[date(year, MAY, 29)] = "Monday before Decoration Day"
            self[date(year, JUL, 3)] = "Monday before Independence Day"
            self[date(year, SEP, 29)] = "Admiral Dewey Celebration"
        elif year == 1900:
            self[date(year, DEC, 24)] = "Christmas Eve"
        elif year == 1901:
            self[date(year, JUL, 5)] = "Friday after Independence Day"
            self[date(year, SEP, 19)] = "Funeral of President McKinley"
        elif year == 1903:
            self[date(year, APR, 22)] = "Opening of new NYSE building"
        elif year == 1914:
            # Beginning of WWI
            begin = date(year, JUL, 31)
            end = date(year, NOV, 27)
            for d in (
                begin + timedelta(days=n)
                for n in range((end - begin).days + 1)
            ):
                if self._is_weekend(d):
                    continue
                self[d] = "World War I"
        elif year == 1917:
            self[date(year, JUN, 5)] = "Draft Registration Day"
        elif year == 1918:
            self[date(year, JAN, 28)] = "Heatless Day"
            self[date(year, FEB, 4)] = "Heatless Day"
            self[date(year, FEB, 11)] = "Heatless Day"
            self[date(year, JUN, 14)] = "Heatless Day"
            self[date(year, SEP, 12)] = "Draft Registration Day"
            self[date(year, NOV, 11)] = "Armistice Day"
        elif year == 1919:
            self[date(year, MAR, 25)] = "Homecoming Day for 27th Division"
            self[date(year, MAY, 6)] = "Parade Day for 77th Division"
            self[date(year, SEP, 10)] = "Return of General Pershing"
        elif year == 1923:
            self[date(year, AUG, 3)] = "Death of President Warren G. Harding"
            self[
                date(year, AUG, 10)
            ] = "Funeral of President Warren G. Harding"
        elif year == 1927:
            self[date(year, JUN, 13)] = "Parade for Colonel Charles Lindbergh"
        elif year == 1929:
            self[date(year, NOV, 29)] = "Catch Up Day"
        elif year == 1933:
            begin = date(year, MAR, 6)
            end = date(year, MAR, 14)
            for d in (
                begin + timedelta(days=n)
                for n in range((end - begin).days + 1)
            ):
                if self._is_weekend(d):
                    continue
                self[d] = "Special Bank Holiday"
        elif year == 1945:
            self[date(year, AUG, 15)] = "V-J Day (WWII)"
            self[date(year, AUG, 16)] = "V-J Day (WWII)"
            self[date(year, DEC, 24)] = "Christmas Eve"
        elif year == 1954:
            self[date(year, DEC, 24)] = "Christmas Eve"
        elif year == 1956:
            self[date(year, DEC, 24)] = "Christmas Eve"
        elif year == 1958:
            self[date(year, DEC, 26)] = "Day after Christmas"
        elif year == 1961:
            self[date(year, MAY, 29)] = "Day before Decoration Day"
        elif year == 1963:
            self[date(year, NOV, 25)] = "Funeral of President John F. Kennedy"
        elif year == 1965:
            self[date(year, DEC, 24)] = "Christmas Eve"
        elif year == 1968:
            self[
                date(year, APR, 9)
            ] = "Day of Mourning for Martin Luther King Jr."
            self[date(year, JUL, 5)] = "Day after Independence Day"
            begin = date(year, JUN, 12)
            end = date(year, DEC, 31)
            for d in (
                begin + timedelta(days=n)
                for n in range((end - begin).days + 1)
            ):
                if d.weekday() != WED:  # Wednesday special holiday
                    continue
                self[d] = "Paper Crisis"
        elif year == 1969:
            self[date(year, FEB, 10)] = "Heavy Snow"
            self[
                date(year, MAR, 31)
            ] = "Funeral of President Dwight D. Eisenhower"
            self[
                date(year, JUL, 21)
            ] = "National Participation in Lunar Exploration"
        elif year == 1972:
            self[date(year, DEC, 28)] = "Funeral for President Harry S. Truman"
        elif year == 1973:
            self[
                date(year, JAN, 25)
            ] = "Funeral for President Lyndon B. Johnson"
        elif year == 1977:
            self[date(year, JUL, 14)] = "Blackout in New Yor City"
        elif year == 1985:
            self[date(year, SEP, 27)] = "Hurricane Gloria"
        elif year == 1994:
            self[
                date(year, APR, 27)
            ] = "Funeral for President Richard M. Nixon"
        elif year == 2001:
            self[date(year, SEP, 11)] = "Closed for Sept 11, 2001 Attacks"
            self[date(year, SEP, 12)] = "Closed for Sept 11, 2001 Attacks"
            self[date(year, SEP, 13)] = "Closed for Sept 11, 2001 Attacks"
            self[date(year, SEP, 14)] = "Closed for Sept 11, 2001 Attacks"
        elif year == 2004:
            self[
                date(year, JUN, 11)
            ] = "Day of Mourning for President Ronald W. Reagan"
        elif year == 2007:
            self[
                date(year, JAN, 2)
            ] = "Day of Mourning for President Gerald R. Ford"
        elif year == 2012:
            self[date(year, OCT, 29)] = "Hurricane Sandy"
            self[date(year, OCT, 30)] = "Hurricane Sandy"
        elif year == 2018:
            self[
                date(year, DEC, 5)
            ] = "Day of Mourning for President George H.W. Bush"


class XNYS(NewYorkStockExchange):
    pass


class NYSE(NewYorkStockExchange):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, APR, MAY, JUN, JUL, AUG, SEP, NOV, DEC
from holidays.constants import FRI, SAT
from holidays.holiday_base import HolidayBase
from holidays.utils import _islamic_to_gre


class UnitedArabEmirates(HolidayBase):

    # Holidays are regulated by the Article 74
    # of Federal Law No. 08 for the year 1980:
    # https://www.ilo.org/dyn/natlex/docs/ELECTRONIC/11956/69376/F417089305/ARE11956.pdf
    # However the law is not applied literally,
    # and was amended often in the past few years.
    # Sources:
    # 2017: https://www.khaleejtimes.com/nation/uae-official-public-holidays-list-2017   # noqa: E501
    # 2018: https://www.thenational.ae/uae/government/uae-public-holidays-2018-announced-by-abu-dhabi-government-1.691393  # noqa: E501
    # 2019: https://www.thenational.ae/uae/government/uae-public-holidays-for-2019-and-2020-announced-by-cabinet-1.833425  # noqa: E501
    # 2020: https://u.ae/en/information-and-services/public-holidays-and-religious-affairs/public-holidays  # noqa: E501

    # Holidays based on the Islamic Calendar are estimated (and so denoted),
    # as are announced each year and based on moon sightings:
    # - Eid al-Fitr*
    # - Eid al-Adha*
    # - Arafat (Hajj) Day*
    # - Al-Hijra (Islamic New Year)*
    # - Mawlud al-Nabi (Prophet Mohammad's Birthday)*
    # - Leilat al-Miraj (Ascension of the Prophet)*, suspended after 2018.
    # *only if hijri-converter library is installed, otherwise a warning is
    #  raised that this holiday is missing. hijri-converter requires
    #  Python >= 3.6
    country = "AE"
    weekend = {FRI, SAT}

    def _populate(self, year):
        super()._populate(year)

        def _add_holiday(dt: date, hol: str) -> None:
            """Only add if in current year; prevents adding holidays across
            years (handles multi-day Islamic holidays that straddle Gregorian
            years).
            """
            if dt.year == year:
                self[dt] = hol

        # New Year's Day
        self[date(year, JAN, 1)] = "New Year's Day"

        # Commemoration Day, since 2015.
        if year >= 2015 and year < 2019:
            self[date(year, NOV, 30)] = "Commemoration Day"
        elif year >= 2019:
            self[date(year, DEC, 1)] = "Commemoration Day"
        else:
            pass

        # National Day
        self[date(year, DEC, 2)] = "National Day"
        self[date(year, DEC, 3)] = "National Day Holiday"

        # Eid al-Fitr
        # Date is announced each year. Usually stretches along 3 or 4 days,
        # in some instances prepending/appending a day or two
        # before/after the official holiday.
        dates_obs = {
            2017: [(JUN, 25)],
            2018: [(JUN, 14)],
            2019: [(JUN, 3)],
            2020: [(MAY, 24)],
        }
        fitr = "Eid al-Fitr"
        if year in dates_obs:
            for date_obs in dates_obs[year]:
                hol_date = date(year, *date_obs)
                self[hol_date] = fitr
                self[hol_date + rd(days=1)] = f"{fitr} Holiday"
                self[hol_date + rd(days=2)] = f"{fitr} Holiday"
        else:
            for yr in (year - 1, year):
                for date_obs in _islamic_to_gre(yr, 10, 1):
                    hol_date = date_obs
                    _add_holiday(hol_date, f"{fitr}* (*estimated)")
                    _add_holiday(
                        hol_date + rd(days=1),
                        f"{fitr} Holiday* (*estimated)",
                    )
                    _add_holiday(
                        hol_date + rd(days=2),
                        f"{fitr} Holiday* (*estimated)",
                    )

        # Arafat Day & Eid al-Adha
        dates_obs = {
            2017: [(AUG, 31)],
            2018: [(AUG, 20)],
            2019: [(AUG, 10)],
            2020: [(JUL, 30)],
        }
        hajj = "Arafat (Hajj) Day"
        adha = "Eid al-Adha"
        if year in dates_obs:
            for date_obs in dates_obs[year]:
                hol_date = date(year, *date_obs)
                self[hol_date] = hajj
                self[hol_date + rd(days=1)] = adha
                self[hol_date + rd(days=2)] = f"{adha} Holiday"
                self[hol_date + rd(days=3)] = f"{adha} Holiday"
        else:
            for yr in (year - 1, year):
                for date_obs in _islamic_to_gre(yr, 12, 9):
                    hol_date = date_obs
                    _add_holiday(hol_date, f"{hajj}* (*estimated)")
                    _add_holiday(
                        hol_date + rd(days=1), f"{adha}* (*estimated)"
                    )
                    _add_holiday(
                        hol_date + rd(days=2),
                        f"{adha}* Holiday* (*estimated)",
                    )
                    _add_holiday(
                        hol_date + rd(days=3),
                        f"{adha} Holiday* (*estimated)",
                    )

        # Islamic New Year - (hijari_year, 1, 1)
        dates_obs = {
            2017: [(SEP, 22)],
            2018: [(SEP, 11)],
            2019: [(AUG, 31)],
            2020: [(AUG, 23)],
        }
        new_hijri_year = "Al Hijra - Islamic New Year"
        if year in dates_obs:
            for date_obs in dates_obs[year]:
                hol_date = date(year, *date_obs)
                self[hol_date] = new_hijri_year
        else:
            for date_obs in _islamic_to_gre(year, 1, 1):
                hol_date = date_obs
                self[hol_date] = f"{new_hijri_year}* (*estimated)"

        # Leilat al-Miraj - The Prophet's ascension (hijari_year, 7, 27)
        if year <= 2018:  # starting from 2019 the UAE government removed this
            dates_obs = {2017: [(APR, 23)], 2018: [(APR, 13)]}
            ascension = "Leilat al-Miraj - The Prophet's ascension"
            if year in dates_obs:
                for date_obs in dates_obs[year]:
                    hol_date = date(year, *date_obs)
                    self[hol_date] = ascension
            else:
                for date_obs in _islamic_to_gre(year, 7, 27):
                    hol_date = date_obs
                    self[hol_date] = f"{ascension}* (*estimated)"

        # Prophet Muhammad's Birthday - (hijari_year, 3, 12)
        if year <= 2019:  # starting from 2020 the UAE government removed this
            dates_obs = {
                2017: [(NOV, 30)],
                2018: [(NOV, 19)],
                2019: [(NOV, 9)],
            }
            mawlud = "Mawlud al-Nabi - Prophet Mohammad's Birthday"
            if year in dates_obs:
                for date_obs in dates_obs[year]:
                    hol_date = date(year, *date_obs)
                    self[hol_date] = mawlud
            else:
                for date_obs in _islamic_to_gre(year, 3, 12):
                    hol_date = date_obs
                    self[hol_date] = f"{mawlud}* (*estimated)"


class AE(UnitedArabEmirates):
    pass


class ARE(UnitedArabEmirates):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)
from datetime import date
from typing import Any

from dateutil.easter import easter
from dateutil.relativedelta import MO
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import MON, JAN, MAR, APR, MAY, JUN, JUL, AUG, SEP
from holidays.constants import NOV, DEC
from holidays.holiday_base import HolidayBase


class UnitedKingdom(HolidayBase):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_the_United_Kingdom
    This class is extended by other countries (Ireland, Isle of Man, ...)
    It must be taken into account when adding or modifying holidays.
    Look at _country_specific() method for country specific behavior.
    """

    country = "GB"
    special_holidays = {
        1977: ((JUN, 7, "Silver Jubilee of Elizabeth II"),),
        1981: ((JUL, 29, "Wedding of Charles and Diana"),),
        1999: ((DEC, 31, "Millennium Celebrations"),),
        2002: ((JUN, 3, "Golden Jubilee of Elizabeth II"),),
        2011: ((APR, 29, "Wedding of William and Catherine"),),
        2012: ((JUN, 5, "Diamond Jubilee of Elizabeth II"),),
        2022: (
            (JUN, 3, "Platinum Jubilee of Elizabeth II"),
            (SEP, 19, "State Funeral of Queen Elizabeth II"),
        ),
        2023: ((MAY, 8, "Coronation of Charles III"),),
    }
    subdivisions = ["UK", "England", "Northern Ireland", "Scotland", "Wales"]

    def __init__(self, **kwargs: Any) -> None:
        # default subdiv to UK; state for backwards compatibility
        if not kwargs.get("subdiv", kwargs.get("state")):
            kwargs["subdiv"] = "UK"
        HolidayBase.__init__(self, **kwargs)

    def _populate(self, year: int) -> None:
        super()._populate(year)

        # New Year's Day
        if year >= 1974:
            name = "New Year's Day"
            dt = date(year, JAN, 1)
            self[dt] = name
            if self.observed and self._is_weekend(dt):
                self[dt + rd(weekday=MO)] = name + " (Observed)"

        # New Year Holiday
        if self.subdiv in {"Scotland", "UK"}:
            name = "New Year Holiday"
            dt = date(year, JAN, 2)
            if self.subdiv == "UK":
                name += " [Scotland]"
            self[dt] = name
            if self.observed and self._is_weekend(dt):
                self[dt + rd(days=+2)] = name + " (Observed)"
            elif self.observed and dt.weekday() == MON:
                self[dt + rd(days=+1)] = name + " (Observed)"

        # St. Patrick's Day
        if self.subdiv in {"Northern Ireland", "UK"}:
            name = "St. Patrick's Day"
            dt = date(year, MAR, 17)
            if self.subdiv == "UK":
                name += " [Northern Ireland]"
            self[dt] = name
            if self.observed and self._is_weekend(dt):
                self[dt + rd(weekday=MO)] = name + " (Observed)"

        # Battle of the Boyne
        if self.subdiv in {"Northern Ireland", "UK"}:
            name = "Battle of the Boyne"
            if self.subdiv == "UK":
                name += " [Northern Ireland]"
            self[date(year, JUL, 12)] = name

        # Summer bank holiday (first Monday in August)
        if self.subdiv in {"Scotland", "UK"}:
            name = "Summer Bank Holiday"
            if self.subdiv == "UK":
                name += " [Scotland]"
            self[date(year, AUG, 1) + rd(weekday=MO)] = name

        # St. Andrew's Day
        if self.subdiv in {"Scotland", "UK"}:
            name = "St. Andrew's Day"
            if self.subdiv == "UK":
                name += " [Scotland]"
            self[date(year, NOV, 30)] = name

        # Christmas Day
        name = "Christmas Day"
        dt = date(year, DEC, 25)
        self[dt] = name
        if self.observed and self._is_weekend(dt):
            self[dt + rd(days=+2)] = name + " (Observed)"

        # Overwrite to modify country specific holidays
        self._country_specific(year)

    def _country_specific(self, year: int) -> None:
        # This method is replaced by class Ireland

        # UnitedKingdom exclusive holidays

        easter_date = easter(year)
        # Good Friday
        self[easter_date + rd(days=-2)] = "Good Friday"

        # Easter Monday
        if self.subdiv != "Scotland":
            name = "Easter Monday"
            if self.subdiv == "UK":
                name += " [England/Wales/Northern Ireland]"
            self[easter_date + rd(days=+1)] = name

        # May Day bank holiday (first Monday in May)
        if year >= 1978:
            name = "May Day"
            if year == 1995 or year == 2020:
                # In 2020 moved to Friday to mark 75th anniversary of VE Day.
                dt = date(year, MAY, 8)
            else:
                dt = date(year, MAY, 1) + rd(weekday=MO)
            self[dt] = name

        # Spring bank holiday (last Monday in May)
        name = "Spring Bank Holiday"
        if year == 2012:
            self[date(year, JUN, 4)] = name
        elif year == 2022:
            self[date(year, JUN, 2)] = name
        elif year >= 1971:
            self[date(year, MAY, 31) + rd(weekday=MO(-1))] = name

        # Late Summer bank holiday (last Monday in August)
        if self.subdiv != "Scotland" and year >= 1971:
            name = "Late Summer Bank Holiday"
            if self.subdiv == "UK":
                name += " [England/Wales/Northern Ireland]"
            self[date(year, AUG, 31) + rd(weekday=MO(-1))] = name

        # Boxing Day
        name = "Boxing Day"
        dt = date(year, DEC, 26)
        self[dt] = name
        if self.observed and self._is_weekend(dt):
            self[dt + rd(days=+2)] = name + " (Observed)"


class UK(UnitedKingdom):
    pass


class GB(UnitedKingdom):
    pass


class GBR(UnitedKingdom):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.relativedelta import relativedelta as rd

from holidays.constants import FEB, SEP, THU, FRI, SAT
from holidays.holiday_base import HolidayBase
from holidays.utils import _islamic_to_gre


class SaudiArabia(HolidayBase):
    """
    There are only 4 official national holidays in Saudi:
    https://laboreducation.hrsd.gov.sa/en/gallery/274
    https://laboreducation.hrsd.gov.sa/en/labor-education/322
    https://english.alarabiya.net/News/gulf/2022/01/27/Saudi-Arabia-to-commemorate-Founding-Day-on-Feb-22-annually-Royal-order
    The national day and the founding day holidays are based on the
    Georgian calendar while the other two holidays are based on the
    Islamic Calendar, and they are estimates as they announced each
    year and based on moon sightings;
    they are:
        - Eid al-Fitr*
        - Eid al-Adha*
    * only if hijri-converter library is installed, otherwise a warning is
    raised that this holiday is missing. hijri-converter requires
    Python >= 3.6
    """

    country = "SA"

    def _populate(self, year):
        super()._populate(year)

        if year < 2013:
            # Weekend used to be THU, FRI before June 28th, 2013
            # On that year both Eids were after that date, and foudning
            # day holiday started at 2022; so what below works,
            self.weekend = (THU, FRI)
        else:
            self.weekend = (FRI, SAT)

        observed_str = " (observed)"

        def _add_holiday(dt: date, hol: str) -> None:
            """Only add if in current year; prevents adding holidays across
            years (handles multi-day Islamic holidays that straddle Gregorian
            years).
            """
            if dt.year == year:
                self[dt] = hol

        # Eid al-Fitr Holiday
        # The holiday is a 4-day holiday starting on the day following the
        # 29th day of Ramadan, the 9th month of the Islamic calendar.
        # Observed days are added to make up for any days falling on a weekend.
        # Holidays may straddle across Gregorian years, so we go back one year
        # to pick up any such occurrence.
        # Date of observance is announced yearly.
        holiday_name = "Eid al-Fitr Holiday"
        for yr in (year - 1, year):
            for hijri_date in _islamic_to_gre(yr, 9, 29):
                hijri_date += rd(days=1)
                for dys in range(4):
                    _add_holiday((hijri_date + rd(days=dys)), holiday_name)
                if self.observed:
                    weekend_days = sum(
                        (hijri_date + rd(days=dys)).weekday() in self.weekend
                        for dys in range(4)
                    )
                    for dys in range(weekend_days):
                        _add_holiday(
                            hijri_date + rd(days=4 + dys),
                            holiday_name + observed_str,
                        )

        # Arafat Day & Eid al-Adha
        # The holiday is a 4-day holiday starting on Arafat Day, the 10th of
        # Dhu al-Hijjah, the 12th month of the Islamic calendar.
        # Observed days are added to make up for any days falling on a weekend.
        # Holidays may straddle across Gregorian years, so we go back one year
        # to pick up any such occurrence.
        # Date of observance is announced yearly.
        holiday_name = "Eid al-Adha Holiday"
        for yr in (year - 1, year):
            for hijri_date in _islamic_to_gre(yr, 12, 9):
                _add_holiday(hijri_date, "Arafat Day Holiday")
                for dys in range(1, 4):
                    _add_holiday((hijri_date + rd(days=dys)), holiday_name)
                if self.observed:
                    weekend_days = sum(
                        (hijri_date + rd(days=dys)).weekday() in self.weekend
                        for dys in range(4)
                    )
                    for dys in range(weekend_days):
                        _add_holiday(
                            hijri_date + rd(days=4 + dys),
                            holiday_name + observed_str,
                        )

        # National Day holiday (started at the year 2005).
        # Note: if national day happens within the Eid al-Fitr Holiday or
        # within Eid al-Fitr Holiday, there is no extra holidays given for it.
        holiday_name = "National Day Holiday"
        if year >= 2005:
            national_day = date(year, SEP, 23)
            if national_day not in self:
                self[national_day] = holiday_name
                # First weekend day(Thursaday before 2013 and Friday otherwise)
                if self.observed and national_day.weekday() == self.weekend[0]:
                    national_day -= rd(days=1)
                    self[national_day] = holiday_name + observed_str
                # Second weekend day(Friday before 2013 and Saturday otherwise)
                elif (
                    self.observed and national_day.weekday() == self.weekend[1]
                ):
                    national_day += rd(days=1)
                    self[national_day] = holiday_name + observed_str

        # Founding Day holiday (started 2022).
        # Note: if founding day happens within the Eid al-Fitr Holiday or
        # within Eid al-Fitr Holiday, there is no extra holidays given for it.
        holiday_name = "Founding Day Holiday"
        if year >= 2022:
            founding_day = date(year, FEB, 22)
            if founding_day not in self:
                self[founding_day] = holiday_name
                # First weekend day(Thursaday before 2013 and Friday otherwise)
                if self.observed and founding_day.weekday() == self.weekend[0]:
                    founding_day -= rd(days=1)
                    self[founding_day] = holiday_name + observed_str
                # Second weekend day(Friday before 2013 and Saturday otherwise)
                elif (
                    self.observed and founding_day.weekday() == self.weekend[1]
                ):
                    founding_day += rd(days=1)
                    self[founding_day] = holiday_name + observed_str


class SA(SaudiArabia):
    pass


class SAU(SaudiArabia):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import EASTER_ORTHODOX, easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, APR, MAY, JUN, JUL, OCT
from holidays.holiday_base import HolidayBase
from holidays.utils import _islamic_to_gre


class Egypt(HolidayBase):

    # Holidays here are estimates, it is common for the day to be pushed
    # if falls in a weekend, although not a rule that can be implemented.
    # Holidays after 2020: the following four moving date holidays whose exact
    # date is announced yearly are estimated (and so denoted):
    # - Eid El Fetr*
    # - Eid El Adha*
    # - Arafat Day*
    # - Moulad El Naby*
    # *only if hijri-converter library is installed, otherwise a warning is
    #  raised that this holiday is missing. hijri-converter requires
    #  Python >= 3.6
    # is_weekend function is there, however not activated for accuracy.

    country = "EG"

    def _populate(self, year):
        super()._populate(year)

        def _add_holiday(dt: date, hol: str) -> None:
            """Only add if in current year; prevents adding holidays across
            years (handles multi-day Islamic holidays that straddle Gregorian
            years).
            """
            if dt.year == year:
                self[dt] = hol

        # New Year's Day
        self[date(year, JAN, 1)] = "New Year's Day - Bank Holiday"

        # Coptic Christmas
        self[date(year, JAN, 7)] = "Coptic Christmas"

        # 25th of Jan
        if year >= 2012:
            self[date(year, JAN, 25)] = "Revolution Day - January 25"
        elif year >= 2009:
            self[date(year, JAN, 25)] = "Police Day"
        else:
            pass

        # Coptic Easter - Orthodox Easter
        easter_date = easter(year, EASTER_ORTHODOX)
        self[easter_date] = "Coptic Easter Sunday"

        # Sham El Nessim - Spring Festival
        self[easter_date + rd(days=+1)] = "Sham El Nessim"

        # Sinai Libration Day
        if year > 1982:
            self[date(year, APR, 25)] = "Sinai Liberation Day"

        # Labour Day
        self[date(year, MAY, 1)] = "Labour Day"

        # Armed Forces Day
        self[date(year, OCT, 6)] = "Armed Forces Day"

        # 30 June Revolution Day
        if year >= 2014:
            self[date(year, JUN, 30)] = "30 June Revolution Day"

        # Revolution Day
        if year > 1952:
            self[date(year, JUL, 23)] = "Revolution Day"

        # Eid al-Fitr - Feast Festive
        # date of observance is announced yearly, This is an estimate since
        # having the Holiday on Weekend does change the number of days,
        # deceided to leave it since marking a Weekend as a holiday
        # wouldn't do much harm.
        for yr in (year - 1, year):
            for date_obs in _islamic_to_gre(yr, 10, 1):
                hol_date = date_obs
                _add_holiday(hol_date, "Eid al-Fitr")
                _add_holiday(hol_date + rd(days=1), "Eid al-Fitr Holiday")
                _add_holiday(hol_date + rd(days=2), "Eid al-Fitr Holiday")

        # Arafat Day & Eid al-Adha - Scarfice Festive
        # date of observance is announced yearly
        for yr in (year - 1, year):
            for date_obs in _islamic_to_gre(yr, 12, 9):
                hol_date = date_obs
                _add_holiday(hol_date, "Arafat Day")
                _add_holiday(hol_date + rd(days=1), "Eid al-Adha")
                _add_holiday(hol_date + rd(days=2), "Eid al-Adha Holiday")
                _add_holiday(hol_date + rd(days=3), "Eid al-Adha Holiday")

        # Islamic New Year - (hijari_year, 1, 1)
        for date_obs in _islamic_to_gre(year, 1, 1):
            hol_date = date_obs
            self[hol_date] = "Islamic New Year"

        # Prophet Muhammad's Birthday - (hijari_year, 3, 12)
        for date_obs in _islamic_to_gre(year, 3, 12):
            hol_date = date_obs
            self[hol_date] = "Prophet Muhammad's Birthday"


class EG(Egypt):
    pass


class EGY(Egypt):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAR, MAY, JUL, AUG, OCT, DEC
from holidays.holiday_base import HolidayBase
from holidays.utils import _islamic_to_gre


class Kazakhstan(HolidayBase):
    """
    1. https://www.officeholidays.com/countries/kazakhstan/2020
    2. https://egov.kz/cms/en/articles/holidays-calend
    3. https://en.wikipedia.org/wiki/Public_holidays_in_Kazakhstan
    4. https://adilet.zan.kz/rus/docs/Z010000267_/history
    """

    country = "KZ"

    def _populate(self, year):
        super()._populate(year)

        # New Year's holiday (2 days)
        name = "New Year"
        self[date(year, JAN, 1)] = name
        self[date(year, JAN, 2)] = name

        # International Women's Day
        self[date(year, MAR, 8)] = "International Women's Day"

        # Nauryz holiday
        name = "Nauryz holiday"
        if year >= 2010:
            self[date(year, MAR, 21)] = name
            self[date(year, MAR, 22)] = name
            self[date(year, MAR, 23)] = name
        elif year >= 2002:
            self[date(year, MAR, 22)] = name

        # Kazakhstan People Solidarity Holiday
        self[date(year, MAY, 1)] = "Kazakhstan People Solidarity Holiday"

        # Defender of the Fatherland Day
        if year >= 2013:
            self[date(year, MAY, 7)] = "Defender of the Fatherland Day"

        # Victory Day
        self[date(year, MAY, 9)] = "Victory Day"

        # Capital Day
        if year >= 2009:
            self[date(year, JUL, 6)] = "Capital Day"

        # Constitution Day of the Republic of Kazakhstan
        if year >= 1996:
            self[
                date(year, AUG, 30)
            ] = "Constitution Day of the Republic of Kazakhstan"

        # Republic Day
        if 1994 <= year <= 2008 or year >= 2022:
            self[date(year, OCT, 25)] = "Republic Day"

        # First President Day
        if 2012 <= year <= 2021:
            self[date(year, DEC, 1)] = "First President Day"

        # Kazakhstan Independence Day
        name = "Kazakhstan Independence Day"
        self[date(year, DEC, 16)] = name
        if 2002 <= year <= 2021:
            self[date(year, DEC, 17)] = name

        if self.observed and year >= 2002:
            for k, v in list(self.items()):
                if self._is_weekend(k) and k.year == year:
                    next_workday = k + rd(days=+1)
                    while self._is_weekend(next_workday) or self.get(
                        next_workday
                    ):
                        next_workday += rd(days=+1)
                    self[next_workday] = v + " (Observed)"

        # Nonworking days (without extending)
        # Orthodox Christmas
        if year >= 2006:
            self[date(year, JAN, 7)] = "Orthodox Christmas"

        # Kurban Ait
        if year >= 2006:
            for hol_date in _islamic_to_gre(year, 12, 10):
                self[hol_date] = "Kurban Ait"


class KZ(Kazakhstan):
    pass


class KAZ(Kazakhstan):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAY, JUL, AUG, NOV, DEC
from holidays.holiday_base import HolidayBase


class Belgium(HolidayBase):
    """
    https://www.belgium.be/nl/over_belgie/land/belgie_in_een_notendop/feestdagen
    https://nl.wikipedia.org/wiki/Feestdagen_in_Belgi%C3%AB
    """

    country = "BE"

    def _populate(self, year):
        super()._populate(year)

        # New years
        self[date(year, JAN, 1)] = "Nieuwjaarsdag"

        easter_date = easter(year)

        # Easter
        self[easter_date] = "Pasen"

        # Second easter day
        self[easter_date + rd(days=+1)] = "Paasmaandag"

        # Ascension day
        self[easter_date + rd(days=+39)] = "O.L.H. Hemelvaart"

        # Pentecost
        self[easter_date + rd(days=+49)] = "Pinksteren"

        # Pentecost monday
        self[easter_date + rd(days=+50)] = "Pinkstermaandag"

        # International Workers' Day
        self[date(year, MAY, 1)] = "Dag van de Arbeid"

        # Belgian National Day
        self[date(year, JUL, 21)] = "Nationale feestdag"

        # Assumption of Mary
        self[date(year, AUG, 15)] = "O.L.V. Hemelvaart"

        # All Saints' Day
        self[date(year, NOV, 1)] = "Allerheiligen"

        # Armistice Day
        self[date(year, NOV, 11)] = "Wapenstilstand"

        # First christmas
        self[date(year, DEC, 25)] = "Kerstmis"


class BE(Belgium):
    pass


class BEL(Belgium):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAY, JUN, OCT, DEC, SUN
from holidays.holiday_base import HolidayBase


class Kenya(HolidayBase):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Kenya
    http://kenyaembassyberlin.de/Public-Holidays-in-Kenya.48.0.html
    https://www.officeholidays.com/holidays/kenya/moi-day
    """

    country = "KE"

    def _populate(self, year):
        super()._populate(year)

        # Public holidays
        self[date(year, JAN, 1)] = "New Year's Day"
        self[date(year, MAY, 1)] = "Labour Day"
        self[date(year, JUN, 1)] = "Madaraka Day"
        self[date(year, OCT, 10)] = "Huduma Day"
        self[date(year, OCT, 20)] = "Mashujaa Day"
        self[date(year, DEC, 12)] = "Jamhuri (Independence) Day"
        self[date(year, DEC, 25)] = "Christmas Day"
        self[date(year, DEC, 26)] = "Utamaduni Day"

        if self.observed:
            for k, v in list(self.items()):
                if k.weekday() == SUN and k.year == year:
                    self[k + rd(days=+1)] = v + " (Observed)"

        easter_date = easter(year)
        self[easter_date + rd(days=-2)] = "Good Friday"
        self[easter_date + rd(days=+1)] = "Easter Monday"


class KE(Kenya):
    pass


class KEN(Kenya):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import SU
from dateutil.relativedelta import relativedelta as rd

from holidays.holiday_base import HolidayBase


class Lithuania(HolidayBase):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Lithuania
    https://www.kalendorius.today/
    """

    country = "LT"

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day
        self[date(year, 1, 1)] = "Naujieji metai"

        # Day of Restoration of the State of Lithuania (1918)
        if year >= 1918:
            self[date(year, 2, 16)] = "Lietuvos valstybės " "atkūrimo diena"

        # Day of Restoration of Independence of Lithuania
        # (from the Soviet Union, 1990)
        if year >= 1990:
            self[date(year, 3, 11)] = (
                "Lietuvos nepriklausomybės " "atkūrimo diena"
            )

        # Easter
        easter_date = easter(year)
        self[easter_date] = "Velykos"

        # Easter 2nd day
        self[easter_date + rd(days=+1)] = "Velykų antroji diena"

        # International Workers' Day
        self[date(year, 5, 1)] = "Tarptautinė darbo diena"

        # Mother's day. First Sunday in May
        self[date(year, 5, 1) + rd(weekday=SU)] = "Motinos diena"

        # Fathers's day. First Sunday in June
        self[date(year, 6, 1) + rd(weekday=SU)] = "Tėvo diena"

        # St. John's Day [Christian name],
        # Day of Dew [original pagan name]
        if year >= 2003:
            self[date(year, 6, 24)] = "Joninės, Rasos"

        # Statehood Day
        if year >= 1991:
            self[date(year, 7, 6)] = (
                "Valstybės (Lietuvos "
                "karaliaus Mindaugo "
                "karūnavimo) diena"
            )

        # Assumption Day
        self[date(year, 8, 15)] = (
            "Žolinė (Švč. Mergelės " "Marijos ėmimo į dangų diena)"
        )

        # All Saints' Day
        self[date(year, 11, 1)] = "Visų šventųjų diena (Vėlinės)"

        # All Souls' Day
        if year >= 2020:
            self[date(year, 11, 2)] = "Mirusiųjų atminimo diena (Vėlinės)"

        # Christmas Eve
        self[date(year, 12, 24)] = "Šv. Kūčios"

        # Christmas 1st day
        self[date(year, 12, 25)] = "Šv. Kalėdų pirma diena"

        # Christmas 2nd day
        self[date(year, 12, 26)] = "Šv. Kalėdų antra diena"


class LT(Lithuania):
    pass


class LTU(Lithuania):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)
from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAR, APR, MAY, JUL, OCT, DEC
from holidays.holiday_base import HolidayBase


class Lesotho(HolidayBase):
    """
    https://tinyurl.com/lesothosmallurl1324251
    https://www.timeanddate.com/holidays/lesotho/
    """

    country = "LS"
    special_holidays = {
        # https://tinyurl.com/lesothourl
        2002: ((APR, 4, "Heroes Day"), (MAY, 25, "Africa Day"))
    }

    def _populate(self, year):
        super()._populate(year)

        if year > 1995:
            # https://www.ilo.org/dyn/travail/docs/2093/Public%20Holidays%20Act%201995.pdf
            self[date(year, JAN, 1)] = "New Year's Day"
            self[date(year, MAR, 11)] = "Moshoeshoe's Day"

            if year < 2002:
                self[date(year, APR, 4)] = "Heroes Day"

            if year > 2002:
                self[date(year, MAY, 25)] = "Africa/Heroes Day"

            easter_date = easter(year)
            self[easter_date + rd(days=-2)] = "Good Friday"
            self[easter_date + rd(days=+1)] = "Easter Monday"
            self[easter_date + rd(days=+39)] = "Ascension Day"
            self[date(year, MAY, 1)] = "Workers' Day"

            if year > 1997:
                # https://en.wikipedia.org/wiki/Letsie_III
                self[date(year, JUL, 17)] = "King's Birthday"
            if year <= 1997:
                self[date(year, MAY, 2)] = "King's Birthday"

            self[date(year, OCT, 4)] = "National Independence Day"
            self[date(year, DEC, 25)] = "Christmas Day"
            self[date(year, DEC, 26)] = "Boxing Day"


class LS(Lesotho):
    pass


class LSO(Lesotho):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAR, APR, MAY, AUG, OCT, NOV, DEC, MON
from holidays.constants import TUE, THU
from holidays.holiday_base import HolidayBase


class Hungary(HolidayBase):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Hungary
    Codification dates:
      - https://hvg.hu/gazdasag/20170307_Megszavaztak_munkaszuneti_nap_lett_a_nagypentek  # noqa
      - https://www.tankonyvtar.hu/hu/tartalom/historia/92-10/ch01.html#id496839
    """

    country = "HU"

    def _populate(self, year: int) -> None:
        super()._populate(year)

        # New year
        self._add_with_observed_day_off(
            date(year, JAN, 1), "Újév", before=False, since=2014
        )

        # National Day
        if 1945 <= year <= 1950 or 1989 <= year:
            self._add_with_observed_day_off(
                date(year, MAR, 15), "Nemzeti ünnep"
            )

        # Soviet era
        if 1950 <= year <= 1989:
            # Proclamation of Soviet socialist governing system
            self[
                date(year, MAR, 21)
            ] = "A Tanácsköztársaság kikiáltásának ünnepe"
            # Liberation Day
            self[date(year, APR, 4)] = "A felszabadulás ünnepe"
            # Memorial day of The Great October Soviet Socialist Revolution
            if year not in {1956, 1989}:
                self[
                    date(year, NOV, 7)
                ] = "A nagy októberi szocialista forradalom ünnepe"

        easter_date = easter(year)

        # Good Friday
        if 2017 <= year:
            self[easter_date + rd(days=-2)] = "Nagypéntek"

        # Easter
        self[easter_date] = "Húsvét"

        # Second easter day
        if 1955 != year:
            self[easter_date + rd(days=+1)] = "Húsvét Hétfő"

        # Pentecost
        self[easter_date + rd(days=+49)] = "Pünkösd"

        # Pentecost monday
        if year <= 1952 or 1992 <= year:
            self[easter_date + rd(days=+50)] = "Pünkösdhétfő"

        # International Workers' Day
        if 1946 <= year:
            self._add_with_observed_day_off(
                date(year, MAY, 1), "A Munka ünnepe"
            )
        if 1950 <= year <= 1953:
            self[date(year, MAY, 2)] = "A Munka ünnepe"

        # State Foundation Day (1771-????, 1891-)
        if 1950 <= year <= 1989:
            self[date(year, AUG, 20)] = "A kenyér ünnepe"
        else:
            self._add_with_observed_day_off(
                date(year, AUG, 20), "Az államalapítás ünnepe"
            )

        # National Day
        if 1991 <= year:
            self._add_with_observed_day_off(
                date(year, OCT, 23), "Nemzeti ünnep"
            )

        # All Saints' Day
        if 1999 <= year:
            self._add_with_observed_day_off(
                date(year, NOV, 1), "Mindenszentek"
            )

        # Christmas Eve is not endorsed officially
        # but nowadays it is usually a day off work
        if (
            self.observed
            and 2010 <= year
            and not self._is_weekend(year, DEC, 24)
        ):
            self[date(year, DEC, 24)] = "Szenteste"

        # First christmas
        self[date(year, DEC, 25)] = "Karácsony"

        # Second christmas
        if 1955 != year:
            self._add_with_observed_day_off(
                date(year, DEC, 26),
                "Karácsony másnapja",
                since=2013,
                before=False,
            )

        # New Year's Eve
        # Since 2014, the last day of the year is an observed day off if New
        # Year's Day falls on a Tuesday.
        if (
            self.observed
            and 2014 <= year
            and date(year, DEC, 31).weekday() == MON
        ):
            self[date(year, DEC, 31)] = "Szilveszter"

    def _add_with_observed_day_off(
        self,
        day: date,
        desc: str,
        since: int = 2010,
        before: bool = True,
        after: bool = True,
    ) -> None:
        # Swapped days off were in place earlier but
        # I haven't found official record yet.
        self[day] = desc
        # TODO: should it be a separate flag?
        if self.observed and since <= day.year:
            if day.weekday() == TUE and before:
                self[day + rd(days=-1)] = desc + " előtti pihenőnap"
            elif day.weekday() == THU and after:
                self[day + rd(days=+1)] = desc + " utáni pihenőnap"


class HU(Hungary):
    pass


class HUN(Hungary):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAY, JUL, AUG, SEP, DEC
from holidays.holiday_base import HolidayBase


class Nicaragua(HolidayBase):
    country = "NI"
    subdivisions = ["MN"]

    def __init__(self, **kwargs):
        # Default subdivision to MN; prov for backwards compatibility
        if not kwargs.get("subdiv", kwargs.get("prov")):
            kwargs["subdiv"] = "MN"
        HolidayBase.__init__(self, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Years
        self[date(year, JAN, 1)] = "Año Nuevo [New Year's Day]"
        # Maundy Thursday
        easter_date = easter(year)
        self[easter_date + rd(days=-3)] = "Jueves Santo [Maundy Thursday]"
        # Good Friday
        self[easter_date + rd(days=-2)] = "Viernes Santo [Good Friday]"
        # Labor Day
        self[date(year, MAY, 1)] = "Día del Trabajo [Labour Day]"
        # Revolution Day
        if year >= 1979:
            self[date(year, JUL, 19)] = "Día de la Revolución [Revolution Day]"
        # Battle of San Jacinto Day
        self[
            date(year, SEP, 14)
        ] = "Batalla de San Jacinto [Battle of San Jacinto]"
        # Independence Day
        self[
            date(year, SEP, 15)
        ] = "Día de la Independencia [Independence Day]"
        # Virgin's Day
        self[date(year, DEC, 8)] = "Concepción de María [Virgin's Day]"
        # Christmas
        self[date(year, DEC, 25)] = "Navidad [Christmas]"

        # Provinces festive day
        # Santo Domingo Day Down
        self[date(year, AUG, 1)] = "Bajada de Santo Domingo"
        # Santo Domingo Day Up
        self[date(year, AUG, 10)] = "Subida de Santo Domingo"


class NI(Nicaragua):
    pass


class NIC(Nicaragua):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from calendar import isleap
from datetime import date

from dateutil.easter import easter, EASTER_ORTHODOX
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAR, MAY, SEP
from holidays.holiday_base import HolidayBase
from holidays.utils import _islamic_to_gre

# Ethiopian holidays are estimated: it is common for the day to be pushed
# if falls in a weekend, although not a rule that can be implemented.
# Holidays after 2020: the following four moving date holidays whose exact
# date is announced yearly are estimated (and so denoted):
# - Eid El Fetr*
# - Eid El Adha*
# - Arafat Day*
# - Moulad El Naby*
# *only if hijri-converter library is installed, otherwise a warning is
#  raised that this holiday is missing. hijri-converter requires
#  Python >= 3.6
# is_weekend function is there, however not activated for accuracy.


class Ethiopia(HolidayBase):

    country = "ET"

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day
        # The Ethiopian New Year is called Kudus Yohannes in Ge'ez and
        # Tigrinya, while in Amharic,
        # the official language of Ethiopia it is called Enkutatash.
        # It occurs on September 11 in the Gregorian Calendar;
        # except for the year preceding a leap year, when it occurs on
        # September 12.
        if self.ethiopian_isleap(year):
            self[date(year, SEP, 12)] = "አዲስ ዓመት እንቁጣጣሽ/Ethiopian New Year"
        else:
            self[date(year, SEP, 11)] = "አዲስ ዓመት እንቁጣጣሽ/Ethiopian New Year"

        # Finding of true cross
        if self.ethiopian_isleap(year):
            self[date(year, SEP, 28)] = "መስቀል/Finding of True Cross"
        else:
            self[date(year, SEP, 27)] = "መስቀል/Finding of True Cross"

        # Ethiopian Christmas
        self[date(year, JAN, 7)] = "ገና/Ethiopian X-Mas"

        # Ethiopian Ephiphany
        self[date(year, JAN, 19)] = "ጥምቀት/Ephiphany"

        # Ethiopian Good Friday
        easter_date = easter(year, EASTER_ORTHODOX)
        self[easter_date + rd(days=-2)] = "ስቅለት/Ethiopian Good Friday"

        # Ethiopian  Easter - Orthodox Easter
        self[easter_date] = "ፋሲካ/Ethiopian Easter"

        # Adwa Victory Day
        if year > 1896:
            self[date(year, MAR, 2)] = "አድዋ/Victory of Adwa"

        # Labour Day
        self[date(year, MAY, 1)] = "የሰራተኞች ቀን/Labor Day"

        # Patriots Day
        if year > 1941:
            self[date(year, MAY, 5)] = "የአርበኞች ቀን/Patriots Day"

        # Derg Downfall Day
        if year > 1991:
            self[
                date(year, MAY, 28)
            ] = "ደርግ የወደቀበት ቀን/Downfall of Dergue regime"

        # Downfall of King. Hailesilassie
        if year < 1991 and year > 1974:
            if self.ethiopian_isleap(year):
                self[date(year, SEP, 13)] = "ደርግ የመጣበት ቀን/Formation of Dergue"
            else:
                self[date(year, SEP, 12)] = "ደርግ የመጣበት ቀን/Formation of Dergue"

        # Eid al-Fitr - Feast Festive
        # date of observance is announced yearly, This is an estimate since
        # having the Holiday on Weekend does change the number of days,
        # deceided to leave it since marking a Weekend as a holiday
        # wouldn't do much harm.
        for date_obs in _islamic_to_gre(year, 10, 1):
            hol_date = date_obs
            self[hol_date] = "ኢድ አልፈጥር/Eid-Al-Fitr"

        # Eid al-Adha - Scarfice Festive
        # date of observance is announced yearly
        for date_obs in _islamic_to_gre(year, 12, 9):
            hol_date = date_obs
            self[hol_date + rd(days=1)] = "አረፋ/Eid-Al-Adha"

        # Prophet Muhammad's Birthday - (hijari_year, 3, 12)
        for date_obs in _islamic_to_gre(year, 3, 12):
            hol_date = date_obs
            self[hol_date + rd(days=1)] = "መውሊድ/Prophet Muhammad's Birthday"

    # Ethiopian leap years are coincident with leap years in the Gregorian
    # calendar until the end of February 2100. It starts earlier from new year
    # of western calendar.
    # Ethiopian leap year starts on Sep 11, so it has an effect on
    # holidays between Sep 11 and Jan 1. Therefore, here on the following
    # function we intentionally add 1 to the leap year to offset the difference
    def ethiopian_isleap(self, year):
        return isleap(year + 1)


class ET(Ethiopia):
    pass


class ETH(Ethiopia):
    pass
